// Brainfuck to LLVM
// Copyright (c) 2004 Marek Materzok
//
// This program is free software; you can redistribute it and/or modify
// it under the terms of the GNU General Public License as published by
// the Free Software Foundation; either version 2 of the License, or 
// (at your option) any later version.

#include <llvm/Module.h>
#include <llvm/DerivedTypes.h>
#include <llvm/Constants.h>
#include <llvm/Instructions.h>
#include <llvm/Bytecode/Writer.h>
#include <llvm/Analysis/LoadValueNumbering.h>
#include <llvm/Transforms/Scalar.h>
#include <llvm/Target/TargetData.h>
#include <llvm/PassManager.h>

#include <iostream>
#include <typeinfo>
#include <vector>
#include <list>
#include <map>

using namespace llvm;

class BfInst;
struct CompilerData;

class BfInstList
{
	public:
		typedef std::list<BfInst *> InstListType;
		typedef InstListType::iterator iterator;
		typedef InstListType::size_type size_type;

	public:
		BfInstList();
		~BfInstList();

		void addInst(BfInst *i) { insts.push_back(i); };

		iterator begin() { return insts.begin(); };
		iterator end() { return insts.end(); };

		size_type size() const { return insts.size(); };

		void clear();

		void compile(CompilerData &c);
	
	private:
		InstListType insts;
};

class BfInst
{
	public:
		BfInst();
		virtual ~BfInst();

		virtual void compile(CompilerData &c) = 0;
};

class BfMoveInst: public BfInst
{
	public:
		BfMoveInst(int pamount): amount(pamount) {};

		int getAmount() const { return amount; };
		void setAmount(int pamount) { amount = pamount; };
		
		virtual void compile(CompilerData &c);

	private:
		int amount;
};

class BfAddInst: public BfInst
{
	public:
		BfAddInst(int pamount): amount(pamount) {};

		int getAmount() const { return amount; };
		void setAmount(int pamount) { amount = pamount; };
		
		virtual void compile(CompilerData &c);

	private:
		int amount;
};

class BfInputInst: public BfInst
{
	public:
		BfInputInst() {};
		
		virtual void compile(CompilerData &c);
};

class BfOutputInst: public BfInst
{
	public:
		BfOutputInst() {};
		
		virtual void compile(CompilerData &c);
};

class BfLoopInst: public BfInst
{
	public:
		BfLoopInst() {};

		BfInstList &getInstList() { return list; };
	
		virtual void compile(CompilerData &c);
		
	private:
		BfInstList list;
};

class BfSimpleLoopInst: public BfInst
{
	public:
		typedef std::pair<int, int> value_type;
		typedef std::map<int, int> ListType;

		BfSimpleLoopInst() {};

		void setMoveAdd(const ListType &ch);

		virtual void compile(CompilerData &c);
		
	private:
		ListType list;
};

struct CompilerData
{
	Function *inputcell;
	Function *outputcell;
	Function *cleararray;
	Value *bfarray;
	Value *bfptr;
	Module *mod;
	BasicBlock *cbl;
	Function *main;
};

class Compiler
{
	public:
		Compiler();

		Module *compile(std::istream &istr);

		static const Type *cell_type;
		static const int bfarray_size;

	private:

		Function *fgetchar;
		Function *fputchar;

		CompilerData cdata;
		BfInstList ilist;

		void declare_externs(Module *mod);
		void declare_helpers(Module *mod);

		void parse(std::istream &istr);
		void parse_list(BfInstList &list, std::string::iterator begin,
				std::string::iterator end);
		void parse_moveadd(BfInstList &list, std::map<int, int> &ch, int end);
};

const Type *Compiler::cell_type = Type::UByteTy;
const int Compiler::bfarray_size = 262144;

BfInstList::BfInstList()
{
}

BfInstList::~BfInstList()
{
	clear();
}

void
BfInstList::compile(CompilerData &c)
{
	for (iterator i = begin(); i != end(); ++i)
	{
		(*i)->compile(c);
	}
}

void
BfInstList::clear()
{
	for (iterator i = begin(); i != end(); ++i)
	{
		delete *i;
	}
	insts.clear();
}

BfInst::BfInst()
{
}

BfInst::~BfInst()
{
}

void
BfMoveInst::compile(CompilerData &c)
{
	if (amount >= 0)
	{
		Value *am = ConstantUInt::get(Type::UIntTy, amount);
		c.bfptr = BinaryOperator::create(Instruction::Add, c.bfptr, am, "", c.cbl);
	}
	else
	{
		Value *am = ConstantUInt::get(Type::UIntTy, -amount);
		c.bfptr = BinaryOperator::create(Instruction::Sub, c.bfptr, am, "", c.cbl);
	}
}

void
BfAddInst::compile(CompilerData &c)
{
	std::vector<Value *> indices;
	indices.push_back(c.bfptr);
	Value *cellptr = new GetElementPtrInst(c.bfarray, indices, "", c.cbl);
	Value *cellval = new LoadInst(cellptr, "", c.cbl);
	if (amount >= 0)
	{
		Value *am = ConstantUInt::get(Compiler::cell_type, amount);
		cellval = BinaryOperator::create(Instruction::Add, cellval, am, "", c.cbl);
	}
	else
	{
		Value *am = ConstantUInt::get(Compiler::cell_type, -amount);
		cellval = BinaryOperator::create(Instruction::Sub, cellval, am, "", c.cbl);
	}
	new StoreInst(cellval, cellptr, c.cbl);
}

void
BfInputInst::compile(CompilerData &c)
{
	std::vector<Value *> indices;
	indices.push_back(c.bfptr);
	Value *cellptr = new GetElementPtrInst(c.bfarray, indices, "", c.cbl);
	Value *val = new CallInst(c.inputcell, std::vector<Value*>(), "", c.cbl);
	new StoreInst(val, cellptr, c.cbl);
}

void
BfOutputInst::compile(CompilerData &c)
{
	std::vector<Value *> indices;
	indices.push_back(c.bfptr);
	Value *cellptr = new GetElementPtrInst(c.bfarray, indices, "", c.cbl);
	Value *cellval = new LoadInst(cellptr, "", c.cbl);
	std::vector<Value*> params;
	params.push_back(cellval);
	new CallInst(c.outputcell, params, "", c.cbl);
}

void
BfLoopInst::compile(CompilerData &c)
{
	if (list.size() == 1 && typeid(**list.begin()) == typeid(BfSimpleLoopInst))
	{
		list.compile(c);
		return;
	}

	BasicBlock *entryBlock = c.cbl;
	BasicBlock *loopBlock = new BasicBlock("", c.main);
	BasicBlock *newBlock = new BasicBlock("", c.main);
	Value *old_bfptr = c.bfptr;

	std::vector<Value *> indices;
	indices.push_back(c.bfptr);
	Value *cellptr = new GetElementPtrInst(c.bfarray, indices, "", c.cbl);
	Value *cellval = new LoadInst(cellptr, "", c.cbl);
	Value *cond = new CastInst(cellval, Type::BoolTy, "", c.cbl);
	new BranchInst(loopBlock, newBlock, cond, c.cbl);
	
	c.cbl = loopBlock;
	PHINode *phinode = new PHINode(Type::UIntTy, "", c.cbl);
	phinode->addIncoming(old_bfptr, entryBlock);
	c.bfptr = phinode;
	list.compile(c);
	BasicBlock *endLoopBlock = c.cbl;
	phinode->addIncoming(c.bfptr, endLoopBlock);
	
	indices.clear();
	indices.push_back(c.bfptr);
	cellptr = new GetElementPtrInst(c.bfarray, indices, "", c.cbl);
	cellval = new LoadInst(cellptr, "", c.cbl);
	cond = new CastInst(cellval, Type::BoolTy, "", c.cbl);
	new BranchInst(loopBlock, newBlock, cond, c.cbl);

	c.cbl = newBlock;
	phinode = new PHINode(Type::UIntTy, "", c.cbl);
	phinode->addIncoming(old_bfptr, entryBlock);
	phinode->addIncoming(c.bfptr, endLoopBlock);
	c.bfptr = phinode;
}
	
void 
BfSimpleLoopInst::setMoveAdd(const ListType &ch)
{
	for (ListType::const_iterator iter = ch.begin(); iter != ch.end(); ++iter)
	{
		if (iter->first == 0) continue;
		list.insert(*iter);
	}
}

void
BfSimpleLoopInst::compile(CompilerData &c)
{
	std::vector<Value *> indices;
	indices.push_back(c.bfptr);
	Value *cellptr = new GetElementPtrInst(c.bfarray, indices, "", c.cbl);
	Value *cellval = new LoadInst(cellptr, "", c.cbl);
	for (ListType::const_iterator iter = list.begin(); iter != list.end(); ++iter)
	{
		Value *am = ConstantUInt::get(Type::UIntTy, iter->first >= 0 ? iter->first : -iter->first);
		Value *nbfptr = BinaryOperator::create(iter->first >= 0 ? Instruction::Add : Instruction::Sub, c.bfptr, am, "", c.cbl);
		indices.clear();
		indices.push_back(nbfptr);
		Value *cell2ptr = new GetElementPtrInst(c.bfarray, indices, "", c.cbl);
		Value *cell2val = new LoadInst(cell2ptr, "", c.cbl);
		am = ConstantUInt::get(Compiler::cell_type, iter->second >= 0 ? iter->second : -iter->second);
		am = BinaryOperator::create(Instruction::Mul, am, cellval, "", c.cbl);
		cell2val = BinaryOperator::create(iter->second >= 0 ? Instruction::Add : Instruction::Sub, cell2val, am, "", c.cbl);
		new StoreInst(cell2val, cell2ptr, c.cbl);
	}
	Value *zero = ConstantUInt::get(Compiler::cell_type, 0);
	new StoreInst(zero, cellptr, c.cbl);
}

Compiler::Compiler()
{
}

void
Compiler::declare_externs(Module *mod)
{
	FunctionType *getchartype = FunctionType::get(Type::IntTy, std::vector<const Type*>(), false);
	fgetchar = new Function(getchartype, Function::ExternalLinkage, "getchar", mod);

	std::vector<const Type *> args;
	args.push_back(Type::IntTy);
	FunctionType *putchartype = FunctionType::get(Type::IntTy, args, false);
	fputchar = new Function(putchartype, Function::ExternalLinkage, "putchar", mod);
}

void
Compiler::declare_helpers(Module *mod)
{
	FunctionType *inputcelltype = FunctionType::get(cell_type, std::vector<const Type*>(), false);
	cdata.inputcell = new Function(inputcelltype, Function::ExternalLinkage, "inputcell", mod);
	
	BasicBlock *entry = new BasicBlock("entry", cdata.inputcell);
	BasicBlock *btrue = new BasicBlock("ok", cdata.inputcell);
	BasicBlock *bfalse = new BasicBlock("eof", cdata.inputcell);
	Value *ch = new CallInst(fgetchar, std::vector<Value*>(), "", entry);
	Value *cond = new SetCondInst(Instruction::SetNE, ch, ConstantSInt::get(Type::IntTy, -1), "", entry);
	new BranchInst(btrue, bfalse, cond, entry);
	
	Value *ret = new CastInst(ch, cell_type, "", btrue);
	new ReturnInst(ret, btrue);

	new ReturnInst(ConstantUInt::get(cell_type, 0), bfalse);
	
	std::vector<const Type *> args;
	args.push_back(cell_type);
	FunctionType *outputcelltype = FunctionType::get(Type::VoidTy, args, false);
	cdata.outputcell = new Function(outputcelltype, Function::ExternalLinkage, "outputcell", mod);

	entry = new BasicBlock("entry", cdata.outputcell);
	Value *arg = cdata.outputcell->abegin();
	ch = new CastInst(arg, Type::IntTy, "", entry);
	std::vector<Value *> params;
	params.push_back(ch);
	ret = new CallInst(fputchar, params, "", entry);

	entry->getInstList().push_back(new ReturnInst());

	args.clear();
	args.push_back(PointerType::get(cell_type));
	FunctionType *cleararraytype = FunctionType::get(Type::VoidTy, args, false);
	cdata.cleararray = new Function(cleararraytype, Function::ExternalLinkage, "cleararray", mod);

	entry = new BasicBlock("entry", cdata.cleararray);
	btrue = new BasicBlock("loop", cdata.cleararray);
	bfalse = new BasicBlock("bye", cdata.cleararray);
	arg = cdata.cleararray->abegin();
	Value *counter = new AllocaInst(Type::UIntTy, 0, "", entry);
	new StoreInst(ConstantUInt::get(Type::UIntTy, 0), counter, entry);
	Value *limit = ConstantUInt::get(Type::UIntTy, bfarray_size);
	new BranchInst(btrue, entry);

	std::vector<Value *> indices;
	Value *counterval = new LoadInst(counter, "", btrue);
	indices.push_back(counterval);
	ch = new GetElementPtrInst(arg, indices, "", btrue);
	new StoreInst(ConstantUInt::get(cell_type, 0), ch, btrue);
	counterval = BinaryOperator::create(Instruction::Add, counterval, ConstantUInt::get(Type::UIntTy, 1), "", btrue);
	new StoreInst(counterval, counter, btrue);
	cond = new SetCondInst(Instruction::SetNE, counterval, limit, "", btrue);
	new BranchInst(btrue, bfalse, cond, btrue);
	
	new ReturnInst(0, bfalse);
}

Module *
Compiler::compile(std::istream &istr)
{
	parse(istr);

	Module *mod = new Module("brainfuck");

	declare_externs(mod);
	declare_helpers(mod);

	FunctionType *maintype = FunctionType::get(Type::IntTy, std::vector<const Type*>(), false);
	cdata.main = new Function(maintype, Function::ExternalLinkage, "main", mod);

	BasicBlock *entry = new BasicBlock("entry", cdata.main);
	Value *bfArraySize = ConstantUInt::get(Type::UIntTy, bfarray_size);
	cdata.bfarray = new MallocInst(cell_type, bfArraySize, "bfarray", entry);
	cdata.bfptr = ConstantUInt::get(Type::UIntTy, 0);
	cdata.cbl = entry;
	cdata.mod = mod;

	std::vector<Value *> params;
	params.push_back(cdata.bfarray);
	new CallInst(cdata.cleararray, params, "", entry);

	ilist.compile(cdata);
	
	new FreeInst(cdata.bfarray, cdata.cbl);
	Value *returnValue = ConstantSInt::get(Type::IntTy, 0);
	new ReturnInst(returnValue, cdata.cbl);
	return mod;
}

void 
Compiler::parse(std::istream &istr)
{
	ilist.clear();

	std::string code;
	while (istr.good())
	{
		char ch = 0;
		istr.get(ch);
		switch(ch)
		{
			case '+': case '-':
			case '<': case '>':
			case '[': case ']':
			case ',': case '.':
				code.push_back(ch);
			default: break;
		}
	}

	parse_list(ilist, code.begin(), code.end());
}

void 
Compiler::parse_list(BfInstList &list, std::string::iterator begin,
				std::string::iterator end)
{
	// TODO: zoptymalizowac!
	std::map<int, int> moveadds;
	int pos = 0;
	for (std::string::iterator i = begin; i != end; ++i)
	{
		char ch = *i;
		switch(ch)
		{
			case '+':
				moveadds[pos]++;
				break;
			case '-':
				moveadds[pos]--;
				break;
			case '>':
				pos++;
				break;
			case '<':
				pos--;
				break;
			case ',':
				parse_moveadd(list, moveadds, pos);
				pos = 0;
				list.addInst(new BfInputInst());
				break;
			case '.':
				parse_moveadd(list, moveadds, pos);
				pos = 0;
				list.addInst(new BfOutputInst());
				break;
			case '[':
			{
				parse_moveadd(list, moveadds, pos);
				pos = 0;
				BfLoopInst *inst = new BfLoopInst();
				std::string::iterator nbegin = ++i;
				int nesting = 0;
				// TODO: optymalizowac
				while (!(*i == ']' && nesting == 0))
				{
					if (*i == '[') nesting++;
					if (*i == ']') nesting--;
					++i;
				}
				parse_list(inst->getInstList(), nbegin, i);
				list.addInst(inst);
				break; 
			}
			case ']':
				// TODO: BLAD
				break;
			default: break;
		};
	}
	// optymalizacja petli [-(blahblah)]
	if (list.size() == 0 && pos == 0 && moveadds[0] == -1)
	{
		BfSimpleLoopInst *inst = new BfSimpleLoopInst();
		inst->setMoveAdd(moveadds);
		list.addInst(inst);
	}
	else
		parse_moveadd(list, moveadds, pos);
}

void 
Compiler::parse_moveadd(BfInstList &list, std::map<int, int> &ch, int end)
{
	int prev = 0;
	for (std::map<int,int>::iterator iter = ch.begin(); iter != ch.end(); ++iter)
	{
		int pos = iter->first - prev;
		int add = iter->second;
		if (add != 0)
		{
			if (pos != 0)
				list.addInst(new BfMoveInst(pos));
			list.addInst(new BfAddInst(add));
			prev = iter->first;
		}
	}
	int pos = end - prev;
	if (pos != 0)
		list.addInst(new BfMoveInst(pos));
	ch.clear();
}

int main(int argc, char **argv)
{
	Compiler comp;

	Module *mod = comp.compile(std::cin);
	llvm::PassManager *passmgr = new llvm::PassManager();
	passmgr->add(new llvm::TargetData("bf2llvm", mod));
	passmgr->add(llvm::createReassociatePass());
	passmgr->add(llvm::createLICMPass());
	passmgr->add(llvm::createLoadValueNumberingPass());
	passmgr->add(llvm::createGCSEPass());
	passmgr->add(llvm::createDeadStoreEliminationPass());
	passmgr->add(llvm::createDeadCodeEliminationPass());
	passmgr->add(llvm::createInstructionCombiningPass());
	passmgr->add(llvm::createLoadValueNumberingPass());
	passmgr->add(llvm::createGCSEPass());
	passmgr->add(llvm::createCFGSimplificationPass());
	passmgr->add(llvm::createLoadValueNumberingPass());
	passmgr->add(llvm::createGCSEPass());
	passmgr->add(llvm::createDeadStoreEliminationPass());

/*
	passmgr->add(llvm::createLoadValueNumberingPass());
	passmgr->add(llvm::createGCSEPass());
	passmgr->add(llvm::createDeadStoreEliminationPass());
	passmgr->add(llvm::createDeadCodeEliminationPass());
	passmgr->add(llvm::createInstructionCombiningPass());
	passmgr->add(llvm::createLoadValueNumberingPass());
	passmgr->add(llvm::createGCSEPass());
	passmgr->add(llvm::createDeadStoreEliminationPass());
	passmgr->add(llvm::createCFGSimplificationPass());
	passmgr->add(llvm::createLoadValueNumberingPass());
	passmgr->add(llvm::createGCSEPass());
	passmgr->add(llvm::createDeadStoreEliminationPass());
*/
	passmgr->run(*mod);

	WriteBytecodeToFile(mod, std::cout);

	delete passmgr;
	delete mod;
	return 0;
}

